#!/bin/bash

APP_CC=mpiicc
APP_CXX=mpiicpc
APP_FC=mpiifort

echo "" 
echo "Building with Intel MPI with Intel Compilers"
echo "============================================"
which ${APP_FC}
if [ $? == 1 ]; then
  echo "ERROR: MPI is not in your path."
  echo "Please load the mpi module and retry."
  exit 1;
fi  

export TOPDIR=${PWD}
export MPI_ROOT=impi

export DEPDIR=${TOPDIR}/ncdeps-$MPI_ROOT/
export DEPBUILDDIR=${TOPDIR}/bdeps-$MPI_ROOT/

mkdir -p $DEPDIR $DEPBUILDDIR
cd $DEPBUILDDIR

echo "Building HDF5"
echo "============="

tar xf ${TOPDIR}/extra/hdf5-1.8.12-mpc.tgz
cd hdf5-1.8.12_MPC
mkdir BUILD
cd BUILD
../configure --enable-parallel --enable-fortran --prefix=${DEPDIR} CC=${APP_CC} FC=${APP_FC} 'CFLAGS=-g -fpic'
make -j 8 || exit 1
make install  || exit 1

echo "Building NetCDF"
echo "==============="

cd $DEPBUILDDIR
tar xf ${TOPDIR}/extra/netcdf-4.4.1.tar.gz
cd netcdf-4.4.1

export CDFFORCE="-I${DEPDIR}/include -L${DEPDIR}/lib -Wl,-rpath,${DEPDIR}/lib"
./configure --prefix=${DEPDIR} CC="${APP_CC} $CDFFORCE" CXX="${APP_CXX} $CDFFORCE" FC="${APP_FC} $CDFFORCE"
make -j8 || exit 1
make install

echo "Building NetCDF Fortran Interface"
echo "================================="
cd $DEPBUILDDIR

tar xf ${TOPDIR}/extra/netcdf-fortran-4.4.4.tar.gz
cd netcdf-fortran-4.4.4
./configure --prefix=${DEPDIR} CC="${APP_CC} $CDFFORCE" CXX="${APP_CXX} $CDFFORCE" FC="${APP_FC} $CDFFORCE"
make -j8 || exit 1
make install || exit 1

echo "Building Homme PIO"
echo "=================="

CODE=$TOPDIR/homme1_3_6/
cd $CODE/utils/pio
make clean || exit 1

cat $TOPDIR/homme1_3_6/utils/pio/Makefile.conf.todo | sed s/mpc_cc/${APP_CC}/g | sed s/mpc_f77/${APP_FC}/g | sed s/mpc_cxx/${APP_CXX}/g | sed s/ncdeps-MPC/ncdeps-${MPI_ROOT}/g > $TOPDIR/homme1_3_6/utils/pio/Makefile.conf

export NETCDF_PATH=$DEPDIR
./configure --enable-netcdf4 NETCDF_PATH=$NETCDF_PATH  CC="$APP_CC ${CDFFORCE}" MPICC="$APP_CC ${CDFFORCE}" MPIF90="$APP_FC  -free ${CDFFORCE}"
make -j8 || exit 1

echo "Building Homme Timing"
echo "====================="

cd $CODE/utils/timing
make clean
/bin/rm -f *.a
make -j8

echo "Building Homme blas"
echo "==================="
cd $CODE/libs/blas
make clean
CC=${APP_CC} MPICC=${APP_CC} MPIF90=${APP_FC} F77=${APP_FC} ./configure
make -j8 

echo "Building Homme lapack"
echo "====================="
cd $CODE/libs/lapack
make clean
CC=${APP_CC} MPICC=${APP_CC} F77=${APP_FC} ./configure
make -j8

echo "Building Homme sweqx"
echo "===================="
cd $CODE/build/sweqx
FC="$APP_FC  -lnetcdff ${CDFFORCE}" CC="$APP_CC ${CDFFORCE}" ./configure --enable-blas --enable-lapack --with-netcdf=$NETCDF_PATH  NP=4 PLEV=1
make clean
make depends CPP="gfortran -E"
make -j8 || exit 1
cp sweqx $TOPDIR/sweqx-$MPI_ROOT


                              
echo "Building Homme preqx"
echo "===================="
cd $CODE/build/preqx
FC="$APP_FC  -lnetcdff ${CDFFORCE}" CC="$APP_CC ${CDFFORCE}" ./configure --enable-blas --enable-lapack --with-netcdf=$NETCDF_PATH  NP=4 PLEV=26 --enable-energy-diagnostics
make clean
make depends CPP="gfortran -E"
make -j8 || exit 1
cp preqx $TOPDIR/preqx-$MPI_ROOT


echo "DONE BUILDING: $MPI_ROOT"
echo "========================"

echo "RUNNING: sweqx-$MPI_ROOT"
echo "========================"
cd $TOPDIR
./run.${MPI_ROOT}.sh
